<?php
declare(strict_types=1);

namespace Inti\Provider;

use Inti\Provider\Query\QueryInterface;
use Inti\Provider\Query\QueryTrait;
use Laminas\Stdlib\ArrayUtils;

use function array_map;
use function array_values;
use function count;

/**
 * Default provider implementation.
 *
 * @package Inti\Provider
 * @author Danijel Fabijan <hipnaba@gmail.com>
 * @link https://bitbucket.org/oktan-nautika/inti-provider
 */
abstract class AbstractProvider implements
    ProviderInterface,
    QueryInterface
{
    use QueryTrait;

    /** @var array<string, string> */
    protected array $field_map = [
        'id' => 'id',
    ];
    /** @var array<string, array<int|string, array>> */
    private array $items_by = [];

    /**
     * Fetches the raw results from the data source.
     *
     * @return array
     */
    abstract protected function fetchResults(): array;

    /**
     * Maps the raw result to the item structure.
     *
     * @param array $result
     * @return array
     */
    protected function mapResultToItem(array $result): array
    {
        return $result;
    }

    /**
     * Maps the raw results array to an array of items.
     *
     * @param array $results
     * @return array
     */
    protected function mapResultsToItems(array $results): array
    {
        $items = array_map([$this, 'mapResultToItem'], $results);

        foreach ($items as $item) {
            $itemId = $this->getId($item);
            $this->items_by['ids'][$itemId] = $item;
        }

        return $items;
    }

    public function getCount(): int
    {
        return count($this->getItems());
    }

    public function getItem(): ?array
    {
        return array_values($this->getItems())[0] ?? null;
    }

    public function getItems(): array
    {
        $this->assertQueryIsInitialized();

        if (isset($this->_query['ids']) && !empty($this->_query['ids'])) {
            return $this->getItemsBy('ids', $this->_query['ids']);
        }

        $results = $this->fetchResults();
        return array_map([$this, 'mapResultToItem'], $results);
    }

    /**
     * @param string $field
     * @param array $values
     * @return array
     */
    private function getItemsBy(string $field, array $values): array
    {
        $local = [];
        $toFetch = [];

        foreach ($values as $value) {
            if (isset($this->items_by[$field][$value])) {
                $local[] = $this->items_by[$field][$value];
            } else {
                $toFetch[] = $value;
            }
        }

        if (count($toFetch) > 0) {
            $fetch = (clone $this);
            $fetch->_query[$field] = $toFetch;

            $fetched = $fetch->fetchResults();
            $fetched = $this->mapResultsToItems($fetched);
        }

        return isset($fetched)
            ? ArrayUtils::merge($local, $fetched ?? [])
            : $local;
    }
}
